ML_FLOW_URI_PATH = (
    "sqlite:///chipiron/scripts/default_output_folder/mlflow_data/mlruns.db"
)

ML_FLOW_URI_PATH_TEST = (
    "sqlite:///chipiron/scripts/default_output_folder/mlflow_data/mlruns_test.db"
)
